const singleNumRegexp = /\w_(\d+)/;
const doubleNumReg = /.*_(\d+)_(\d+)$/;

class Base {
    static parseID($this) {
        singleNumRegexp.lastIndex = 0;
        let found = singleNumRegexp.exec($this.attr('id'));
        return found ? parseInt(found[1]) : 0;
    }

    static parseClass($this) {
        singleNumRegexp.lastIndex = 0;
        let found = singleNumRegexp.exec($this.attr('class'));
        return found ? parseInt(found[1]) : 0;
    }

    static parsePosition($this) {
        doubleNumReg.lastIndex = 0;
        let [col, row] = doubleNumReg.exec($this.attr('id')).splice(1, 2);
        return [ parseInt(col), parseInt(row) ];
    }

    static parseSettings() {
        let $hiddenInfo = $('#hidden_div');
        let $pageType = $('#pageType');

        let settings = $.parseJSON($hiddenInfo.html());
        settings.disciplineID = settings.ID;
        settings.pageType = $pageType.val();

        $hiddenInfo.remove();
        $pageType.remove();

        return settings;
    }
}

class Cursor {
    constructor() {
        this.col = 0;
        this.row = 0;
    }

    update(col, row) {
        this.col = col;
        this.row = row;
    }

    get() {
        return [this.col, this.row];
    }

    parseLocation($this) {
        let [col, row] = Base.parsePosition($this);
        this.update(col, row);
    }

    isEmpty() {
        return (this.col == 0) && (this.row == 0);
    }
}

class Cell {
    constructor() {
        this.update(0, 0, 0);
    }

    update(student, submodule, maxRate) {
        this.student = student;
        this.submodule = submodule;
        this.maxRate = maxRate;
    }
}

class CellInfo {
    constructor(cursor, cell) {
        this.cursor = cursor;
        this.cell = cell;

        this.$modulesHead = $('#modulesHead');
        this.$submodulesHead = $('#submodulesHead');
        this.$submodulesMaxRate = $('#submodulesHeadMaxRate');
    }

    select($this) {
        this.cursor.parseLocation($this);

        let [col] = this.cursor.get();

        let $curModuleHead = this.$modulesHead.children(`.col_${col}`);
        let $curSubmoduleHead = this.$submodulesHead.children(`.col_${col}`);
        let $curSubmoduleMaxRate = this.$submodulesMaxRate.children(`.col_${col}`);

        let student = Base.parseID($this.siblings('.name'));
        let submodule = Base.parseID($curSubmoduleMaxRate);
        let maxRate = parseInt($curSubmoduleMaxRate.text());
        if ($this.hasClass('extra')) {
            maxRate -= parseInt($this.siblings('.regular').text());
            maxRate -= parseInt($this.prev('.extra').text()) || 0;
            maxRate = Math.max(0, maxRate);
        }
        this.cell.update(student, submodule, maxRate);
    }

    reset() {
        this.cell.update(0, 0, 0);
    }
}

class RateInfo {
    constructor(cursor, cell) {
        this.cursor = cursor;
        this.cell = cell;

        this.$modulesHead = $('#modulesHead');
        this.$submodulesHead = $('#submodulesHead');
        this.$submodulesMaxRate = $('#submodulesHeadMaxRate');

        this.$cellInfo = $('#cellInfo');
        this.$studentInfo = this.$cellInfo.find('#student .value');
        this.$submoduleInfo = this.$cellInfo.find('#submodule .value');
        this.$rateInfo = this.$cellInfo.find('#maxRate .value');
    }

    show($this) {
        let [col] = this.cursor.get();

        let studentName = $this.siblings('.name').text();
        let submoduleTitle = this.$submodulesHead.children(`.col_${col}`).text();

        this.$studentInfo.html(studentName);
        this.$submoduleInfo.html(submoduleTitle);
        this.$rateInfo.html(this.cell.maxRate);
        this.$cellInfo.show();
    }

    hide() {
        this.$cellInfo.hide();
    }
}

class GroupFilter {
    constructor(settings, $table, $selector) {
        this.settings = settings;
        this.$table = $table.children('tbody');

        let groupFilter = JSON.parse(localStorage.groupFilter || "{}");
        let groupID = groupFilter[this.settings.disciplineID] || 0;

        $selector.children('[value=' + groupID + ']').prop('selected', true);
        this.filterGroups(groupID);

        this.listen($selector);
    }

    filterGroups(groupID) {
        let $trs = this.$table.children();
        $trs.show();

        if (groupID) {
            $trs.not(`.group_${groupID}`).hide();
        }
    }

    listen($selector) {
        $selector.change(() => {
            let groupID = parseInt(parseInt($selector.val()));

            let groupFilter = JSON.parse(localStorage.groupFilter || "{}");
            groupFilter[this.settings.disciplineID] = groupID;
            localStorage.groupFilter = JSON.stringify(groupFilter);

            this.filterGroups(groupID);
        });
    }
}

class Rating {
    constructor() {
        this.settings = Base.parseSettings();
        this.cell = new Cell();
        this.cursor = new Cursor();
        this.cellInfo = new CellInfo(this.cursor, this.cell);
        this.rateInfo = new RateInfo(this.cursor, this.cell);
        this.filter = new GroupFilter(this.settings, $('#studentsRate'), $('#groupSelector'));

        this.Direction = {
            Up: 0,
            Right: 1,
            Down: 2,
            Left: 3
        };

        let keyDirs = {
            13: this.Direction.Down,  // enter
            40: this.Direction.Down,  // down arrow
            38: this.Direction.Up,    // up arrow
            39: this.Direction.Right, // right arrow
            37: this.Direction.Left   // left arrow
        };

        let self = this;
        let oldInputVal = '';
        $('#studentsRate tbody').on('focusin', '.rate', function () {
            self.cellInfo.select($(this));
            self.rateInfo.show($(this));
            $(this).parents('tr').addClass('focus');
            $(this).children('input').select()

            oldInputVal = $(this).children('input').val();
        }).on('focusout', '.rate', function () {
            self.rateInfo.hide($(this));
            self.cellInfo.reset();
            $(this).parents('tr').removeClass('focus');
        }).on('keydown', '.rate', function (event) {
            let key = event.keyCode;

            if (key === 27) {
                $(this).children('input').val(oldInputVal);
                $(this).children('input').blur();
            }

            if (!(key in keyDirs)) {
                return;
            }

            let [col, row] = self.cursor.get();
            let direction = keyDirs[key];
            let $focus = self.getDesiredCell(direction);

            if ($focus.length) {
                $focus.children('input').focus();
            } else {
                $(this).children('input').blur();
            }
        }).on('change', '.rate', function () {
            // Фильтрация символов кроме чисел
            let scoreInputVal = $(this).children('input').val();
            var numbers = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9'];
            scoreInputVal = Array.prototype.filter.call(scoreInputVal, (chr) => chr in numbers).join('');
            $(this).children('input').val(scoreInputVal);

            if (scoreInputVal === oldInputVal) {
                return;
            }

            self.setRate($(this), scoreInputVal, oldInputVal);
        }).on('change', '.checkbox', function () {
            var option;
            if ($(this).hasClass('absence')) option = 'absence';
            if ($(this).hasClass('autopass')) option = 'pass';
            if (!$(this).children('input').prop('checked')) option = null;

            self.cellInfo.select($(this));
            self.setExam($(this), option);
            self.cellInfo.reset();
        }).on('click', '.download button', function () {
            $.fileDownload(URLdir + 'handler/FileCreator/GenerateFinalForm', {
                httpMethod: 'POST',
                data: {
                    'disciplineID': self.settings.disciplineID,
                    'groupID': Base.parseID($(this)),
                    'stage': parseInt($(this).siblings('select').val()),
                },
            });
        });
    }

    /**
     * @param direction Direction в каком направлении искать следующую ячейку для перемещения фокуса
     * @return jQuery клетка, в которую надо переместиться или пустой jquery list, если в этом направлении нет подходящих ячеек
     */
    getDesiredCell(direction) {
        let [col, row] = this.cursor.get();

        let $cell;
        do {
            if (direction == this.Direction.Left)  col--;
            if (direction == this.Direction.Right) col++;
            if (direction == this.Direction.Up)    row--;
            if (direction == this.Direction.Down)  row++;
            $cell = $(`#col_row_${col}_${row}`);
        } while ($cell.length && (!$cell.hasClass('rate') || $cell.children('input').prop('disabled')));

        return $cell;
    };

    setRate($this, scoreInputVal, oldInputVal) {
        let $input = $this.children('input');

        $input.turnOff();

        let studentID = this.cell.student;
        let disciplineID = this.settings.disciplineID;
        let submoduleID = this.cell.submodule;
        let rate = parseInt(scoreInputVal || -1);
        let [col, row] = this.cursor.get();

        $.postJSON(URLdir + 'handler/rating/setRate', {
            studentID, disciplineID, submoduleID, rate
        }).success(() => {
            let rateResult = this.recountScores($this);
            let rateResultText = rateResult;
            if (rateResult < 0) rateResultText = '';
            if (rateResult > 100) rateResultText = '100+';
            $this.siblings('.result').text(rateResultText);

            if ($this.hasClass('extra')) {
                let $cell = $this.next('.extra');
                if ($cell.length) {
                    let bonusRate = this.parseRate($this.siblings('.bonus'));
                    let maxVal = Math.max(60 - rateResult + bonusRate, 0);
                    let placeholder = (maxVal > 0 && rate != -1) ? 'макс. ' + maxVal : '–';
                    $cell.find('input').attr('placeholder', placeholder);
                }
            }

            if ($this.hasClass('exam')) {
                let absenceCol = Base.parseClass($(`#absence_${submoduleID}`));
                let autoPassCol = Base.parseClass($(`#autopass_${submoduleID}`));

                $(`#col_row_${absenceCol}_${row} input`).prop('checked', false);
                $(`#col_row_${autoPassCol}_${row} input`).prop('checked', false);
            }

            Popup.success('Балл добавлен/изменен');
        }).fail((jqXHR) => {
            $this.children('input').val(oldInputVal);
            let status = parseInt(jqXHR.status);
            switch (status) {
                case 400: Popup.error(JSON.parse(jqXHR.responseText || '{}').message); break;
                case 403: Popup.error('Сессия истекла'); break;
                default:  Popup.error('Не удалось добавить/изменить балл');
            }

            if (status == 403) {
                window.location.replace(URLdir);
            }
        }).always(function () {
            $input.turnOn();
        });
    }

    setExam($this, option) {
        let $input = $this.children('input');

        $input.turnOff();

        let studentID = this.cell.student;
        let disciplineID = this.settings.disciplineID;
        let submoduleID = this.cell.submodule;
        option = String(option);
        let [col, row] = this.cursor.get();

        $.postJSON(URLdir + 'handler/rating/SetExamPeriodOption', {
            studentID, disciplineID, submoduleID, option
        }).done(data => {
            if ($this.hasClass('absence')) {
                let examCol = Base.parseClass($(`#exam_${submoduleID}`));
                let autoPassCol = Base.parseClass($(`#autopass_${submoduleID}`));

                $(`#col_row_${examCol}_${row} input`).val('');
                $(`#col_row_${autoPassCol}_${row} input`).prop('checked', false);
            }

            if ($this.hasClass('autopass')) {
                let examCol = Base.parseClass($(`#exam_${submoduleID}`));
                let absenceCol = Base.parseClass($(`#absence_${submoduleID}`));

                $(`#col_row_${examCol}_${row} input`).val('');
                $(`#col_row_${absenceCol}_${row} input`).prop('checked', false);
            }

            Popup.success('Значение установлено');
        }).fail(jqXHR => {
            $input.prop("checked", !$input.prop("checked"));
            let status = parseInt(jqXHR.status);
            switch (status) {
                case 400: Popup.error(JSON.parse(jqXHR.responseText || '{}').message); break;
                case 403: Popup.error('Сессия истекла'); break;
                default:  Popup.error('Не удалось установить значение');
            }

            if (status === 403) {
                window.location.replace(URLdir);
            }
        }).always(function () {
            $input.turnOn();
        });
    }

    recountScores($this) {
        let $rateCells = $this.parent().children('.rate').not('.result');

        let rateResult = 0;
        rateResult += $rateCells.not('.exam').aggregate((init, elem) => init + this.parseRate($(elem)));
        rateResult += $rateCells.filter('.exam').aggregate((init, elem) => Math.max(init, this.parseRate($(elem))));

        return rateResult;
    }

    parseRate($cell) {
        if ($cell.hasClass('static')) {
            return parseInt($cell.text()) | 0;
        } else {
            return parseInt($cell.children('input').val()) | 0;
        }
    }
}

$(() => new Rating());
